//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
#include "gnss/capture/capture_gnss.h"
#include "gnss/capture/capture_gnss_fix.h"
#include "gnss/factor/factor_gnss_fix_2d.h"
#include "gnss/factor/factor_gnss_fix_3d.h"
#include "gnss/feature/feature_gnss_fix.h"
#include "gnss/processor/processor_gnss_fix.h"

namespace wolf
{

ProcessorGnssFix::ProcessorGnssFix(ParamsProcessorGnssFixPtr _params_gnss) :
        ProcessorBase("ProcessorGnssFix", 0, _params_gnss),
        params_gnss_(_params_gnss)
{
    //
}

ProcessorGnssFix::~ProcessorGnssFix()
{
    //
}

void ProcessorGnssFix::processCapture(CaptureBasePtr _capture)
{
    // TODO: keep captures in a buffer and deal with KFpacks
    WOLF_DEBUG("PR ", getName()," process() capture type: ", _capture->getType());
    incoming_capture_ = _capture;

    bool KF_created = false;

    // Check type of capture: GNSS raw
    bool israw = false;
    if (std::dynamic_pointer_cast<CaptureGnss>(_capture)!=nullptr)
        israw = true;
    // if not raw nor Fix, not processable
    else if (std::dynamic_pointer_cast<CaptureGnssFix>(_capture)==nullptr)
        return;

    // Only process raw if fix_from_raw
    if (israw and !params_gnss_->fix_from_raw)
        return;

    // Only process fix if not fix_from_raw
    if (!israw and params_gnss_->fix_from_raw)
        return;

    FrameBasePtr new_frame = nullptr;
    FactorBasePtr new_fac = nullptr;

    // emplace features
    if (israw)
    {
        auto raw_capture = std::static_pointer_cast<CaptureGnss>(incoming_capture_);
        incoming_pos_out_ = GnssUtils::computePos(*(raw_capture->getObservations()), *(raw_capture->getNavigation()), params_gnss_->compute_pos_opt);
        if (!incoming_pos_out_.success) // error
        {
            WOLF_DEBUG("computePos failed with msg: ", incoming_pos_out_.msg);
            return;
        }
    }
    else
    {
        WOLF_DEBUG("ProcessorGnssFix: creating feature from incoming navsatfix. Type:", incoming_capture_->getType());
        auto fix_capture = std::static_pointer_cast<CaptureGnssFix>(incoming_capture_);
        incoming_pos_out_.time = fix_capture->getGpsTimeStamp().getSeconds();           // time_t time;
        incoming_pos_out_.sec =  1e-9 * fix_capture->getGpsTimeStamp().getSeconds();    // double sec;
        incoming_pos_out_.pos = fix_capture->getPositionEcef();                         // Eigen::Vector3d pos;        // position (m)
        incoming_pos_out_.vel = Eigen::Vector3d::Zero();                                // Eigen::Vector3d vel;        // velocity (m/s)
        incoming_pos_out_.pos_covar = fix_capture->getCovarianceEcef();                 // Eigen::Matrix3d pos_covar;  // position covariance (m^2)
        incoming_pos_out_.type = 0;                                                     // int type;                   // coordinates used (0:xyz-ecef,1:enu-baseline)
    }
    auto incoming_feature = FeatureBase::emplace<FeatureGnssFix>(incoming_capture_, incoming_pos_out_);

    // ALREADY CREATED KF
    FrameBasePtr keyframe = buffer_frame_.select( incoming_capture_->getTimeStamp(), params_gnss_->time_tolerance);
    if (keyframe and last_KF_capture_ and keyframe == last_KF_capture_->getFrame())
        keyframe = nullptr;
    if (keyframe)
    {
        WOLF_DEBUG("PR ", getName()," - capture ", incoming_capture_->id(), " appended to existing KF " , keyframe->id() , " TS: ", keyframe->getTimeStamp());
        new_frame = keyframe;
    }
    // MAKE KF
    else if (permittedKeyFrame() && voteForKeyFrame())
    {
        WOLF_DEBUG("PR ", getName()," emplacing KF TS = ", incoming_capture_->getTimeStamp());
        new_frame = getProblem()->emplaceFrame(incoming_capture_->getTimeStamp());
        KF_created = true;
    }
    // OTHERWISE return
    else
        return;

    // ESTABLISH FACTOR
    // link capture
    incoming_capture_->link(new_frame);
    // emplace factor
    new_fac = emplaceFactor(incoming_feature);

    // outlier rejection (only can be evaluated if ENU defined and ENU-MAP well initialized: fixed)
    if (params_gnss_->remove_outliers and
        sensor_gnss_->isEnuDefined() and
        sensor_gnss_->isEnuMapFixed() and
        detectOutlier(new_fac))
    {
        new_frame->remove();
        return;
    }

    // store last KF
    last_KF_capture_ = incoming_capture_;
    last_KF_feature_ = incoming_feature;

    // Define ENU (if not defined)
    if (!sensor_gnss_->isEnuDefined())
    {
        WOLF_DEBUG("Defining ecef enu");
        sensor_gnss_->setEcefEnu(incoming_feature->getMeasurement());
    }

    // Store the first capture that established a factor (for initializing ENU-MAP)
    if (first_frame_state_.empty() and
        not sensor_gnss_->isEnuMapFixed())
    {
        first_frame_state_ = incoming_capture_->getFrame()->getState();
        first_pos_ = incoming_feature->getMeasurement();
    }
    // Initialize ENU-MAP if: ENU defined and ENU-MAP not initialized (and not fixed) and far enough
    if (params_gnss_->init_enu_map and
        not first_frame_state_.empty() and
        sensor_gnss_->isEnuDefined() and
        not sensor_gnss_->isEnuMapInitialized() and
        not sensor_gnss_->isEnuMapFixed() and
        (first_pos_-incoming_pos_out_.pos).norm() > params_gnss_->enu_map_init_dist_min)
    {
        assert(first_frame_state_.count('P') and first_frame_state_.count('O') and incoming_capture_->getFrame() != nullptr);

        sensor_gnss_->initializeEnuMap(first_frame_state_.vector("PO"),
                                       first_pos_,
                                       incoming_capture_->getFrame()->getState().vector("PO"),
                                       incoming_feature->getMeasurement());
        // Set as not-initialized if factors not separated enough ( < enu_map_init_dist_max)
        if ((first_pos_ - incoming_feature->getMeasurement()).norm() < params_gnss_->enu_map_init_dist_max)
            sensor_gnss_->setEnuMapInitialized(false);
    }

    // Fix ENU-MAP if separated enough from first pos ( > enu_map_fix_dist )
    if ((first_pos_ - incoming_feature->getMeasurement()).norm() > params_gnss_->enu_map_fix_dist)
        sensor_gnss_->fixEnuMap();

    // Notify if KF created
    if (KF_created)
        getProblem()->keyFrameCallback(new_frame, shared_from_this());
}

FactorBasePtr ProcessorGnssFix::emplaceFactor(FeatureBasePtr _ftr)
{
    //WOLF_DEBUG("creating the factor...");
    // 2d
    if (getProblem()->getDim() == 2)
        return FactorBase::emplace<FactorGnssFix2d>(_ftr, _ftr, sensor_gnss_, shared_from_this(), params_->apply_loss_function, FAC_ACTIVE);
    // 3d
    else
        return FactorBase::emplace<FactorGnssFix3d>(_ftr, _ftr, sensor_gnss_, shared_from_this(), params_->apply_loss_function, FAC_ACTIVE);
}

bool ProcessorGnssFix::voteForKeyFrame() const
{
    //WOLF_DEBUG("voteForKeyFrame...");
    // Null lastKF
    if (not last_KF_capture_)
    {
        WOLF_DEBUG("KF because of null lastKF");
        return true;
    }

    // Depending on time since the last KF with gnssfix capture
    if (not last_KF_capture_ or
        (incoming_capture_->getTimeStamp() - last_KF_capture_->getTimeStamp()) > params_gnss_->max_time_span)
    {
        WOLF_DEBUG("KF because of time since last KF");
        return true;
    }

    // ENU not defined
    if (not sensor_gnss_->isEnuDefined())
    {
        WOLF_DEBUG("KF because of enu not defined");
        return true;
    }

    // ENU-MAP not initialized and can be initialized
    if (params_gnss_->init_enu_map and
        not first_frame_state_.empty() and
        sensor_gnss_->isEnuDefined() and
        not sensor_gnss_->isEnuMapInitialized() and
        not sensor_gnss_->isEnuMapFixed() and
        (first_pos_-incoming_pos_out_.pos).norm() > params_gnss_->enu_map_init_dist_min and
        (first_pos_-incoming_pos_out_.pos).norm() < params_gnss_->enu_map_init_dist_max)
    {
        WOLF_DEBUG("KF because of enu map not initialized");
        return true;
    }

    // Distance criterion (ENU defined and ENU-MAP initialized)
    if (last_KF_capture_ != nullptr and
        (incoming_pos_out_.pos - last_KF_feature_->getMeasurement()).norm() > params_gnss_->dist_traveled)
    {
        WOLF_DEBUG("KF because of distance criterion: ", (incoming_pos_out_.pos - last_KF_feature_->getMeasurement()).norm());
        return true;
    }

    // otherwise
    return false;
}

bool ProcessorGnssFix::detectOutlier(FactorBasePtr fac)
{
    WOLF_DEBUG( "PR ", getName()," rejectOutlier...");

    // Cast feature
    auto gnss_ftr = std::static_pointer_cast<FeatureGnssFix>(fac->getFeature());

    // copy states
    Eigen::VectorXd x(gnss_ftr->getCapture()->getFrame()->getP()->getState());
    Eigen::VectorXd o(gnss_ftr->getCapture()->getFrame()->getO()->getState());
    Eigen::VectorXd x_antena(sensor_gnss_->getP()->getState());
    Eigen::VectorXd t_ENU_map(sensor_gnss_->getEnuMapTranslation()->getState());
    Eigen::VectorXd roll_ENU_map(sensor_gnss_->getEnuMapRoll()->getState());
    Eigen::VectorXd pitch_ENU_map(sensor_gnss_->getEnuMapPitch()->getState());
    Eigen::VectorXd yaw_ENU_map(sensor_gnss_->getEnuMapYaw()->getState());

    // create double* array of a copy of the state
    double* parameters[7] = {x.data(), o.data(), x_antena.data(), t_ENU_map.data(), roll_ENU_map.data(),
                             pitch_ENU_map.data(), yaw_ENU_map.data()};
    // residual
    Eigen::Vector3d residual;

    // evaluate the factor in this state
    fac->evaluate(parameters, residual.data(), nullptr);

    // evaluate if residual too high at the current estimation
    if (residual.norm() > params_gnss_->outlier_residual_th)
    {
        WOLF_WARN("Discarding GNSS FIX Factor, considered OUTLIER");
        WOLF_TRACE("Feature: ", fac->getMeasurement().transpose(),
                   "\n\tError: ",(fac->getMeasurementSquareRootInformationUpper().inverse()*residual).transpose(),
                   "\n\tResidual: ",residual.transpose(),
                   "\n\tResidual norm: ",residual.norm(), "(max: ", params_gnss_->outlier_residual_th, ")");
        return true;
    }
    return false;
}

bool ProcessorGnssFix::storeKeyFrame(FrameBasePtr _frame_ptr)
{
    return true;
}
bool ProcessorGnssFix::storeCapture(CaptureBasePtr _cap_ptr)
{
    return false;
}
void ProcessorGnssFix::configure(SensorBasePtr _sensor)
{
    sensor_gnss_ = std::static_pointer_cast<SensorGnss>(_sensor);
};

} // namespace wolf

// Register in the FactoryProcessor
#include "core/processor/factory_processor.h"
namespace wolf {
WOLF_REGISTER_PROCESSOR(ProcessorGnssFix)
WOLF_REGISTER_PROCESSOR_AUTO(ProcessorGnssFix);
} // namespace wolf
